
/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

/** \file sol.h
    \brief Header for the generic Solver.
*/


#ifndef _SOLVER_H
#define _SOLVER_H

#include <mistral_mod.h>
#include <mistral_glo.h>
#include <string>
#include <stdexcept>

//using namespace std;

/**********************************************************************/

namespace Mistral {

  class SolutionMethod {

  public:

    Solver *solver;

    SolutionMethod(Solver *s) { solver = s; }
    virtual ~SolutionMethod() {}
  
    virtual void execute() = 0;
    virtual void initialise() = 0;
  };


  class abort_search : public std::exception {
  public:
    abort_search() : exception() {}
  };


  /**********************************************
   * MistralGacVarStack
   **********************************************/

  /// Stack of Variable for computing the GAC closure.
  class MistralGacVarStack {
  public:
    /**@name Parameters*/
    //@{
    VariableInt **MistralGacVarStack_;
    int *trigger;
    int     size;
    //@}
  
    /**@name Constructors*/
    //@{
    MistralGacVarStack()
    {
      size=0;
    }
    MistralGacVarStack(int n)
    {
      initStack(n);
    }
    ~MistralGacVarStack()
    {
      delete [] trigger;
      delete [] MistralGacVarStack_;
    }
    void initStack(int n)
    { 
      trigger = new int[n];
      std::fill( trigger, trigger+n, 0 );
      MistralGacVarStack_ = new VariableInt*[n];
      size = 0;
    }
    //@}
  
    /**@name Accessors*/
    //@{
    inline void push(VariableInt* v, int event)
    {
      if( !trigger[v->id] )
	MistralGacVarStack_[size++] = v;    
      trigger[v->id] |= event;
    }

    inline void push(VariableInt* v, VariableInt* x)
    {
      if( !trigger[v->id] )
	MistralGacVarStack_[size++] = v;    
      trigger[v->id] = x->id;
    }

    inline VariableInt* pop( int& event )
    {
      int j, i = size, mindom = NOVAL;
      VariableInt *best, *last = MistralGacVarStack_[--size];
      while( i-- )
	if( mindom > MistralGacVarStack_[i]->domsize() ) 
	  {
	    j = i;
	    mindom = MistralGacVarStack_[i]->domsize();
	  }
      event = trigger[MistralGacVarStack_[j]->id];
      trigger[MistralGacVarStack_[j]->id] = 0;
      best = MistralGacVarStack_[j];
      MistralGacVarStack_[j] = last;
      return best;
    }

    inline VariableInt* back( int& event )
    {
      VariableInt* last = MistralGacVarStack_[size-1];
      event = trigger[last->id];
      return last;
    }

    inline bool empty() const
    {
      return (!size);
    }

    inline void clear()
    {
      while( size )
	trigger[MistralGacVarStack_[--size]->id] = 0; 
    }
    //@}
  };



  /**********************************************
   * Solver
   **********************************************/

  /// A representation of the search process.
  /**
     Maintains data structures 
     and Implements the generic backtracking method
  */
  class Solver {

  public:
    /*!@name Static Parameters*/
    //@{
    /// Whether the constraint are optimized for bound or ac reasoning
    bool boundOptimised;
    /// Whether we use domain splitting as branching strategy
    bool domainSplitting;
    /// Whether the variables are randomly shuffled before a restart
    bool randomizedRestart;

    /// A method to execute whenever a solution is found (optional)
    SolutionMethod *function;
    /// The objective function, for optimisation problem (optional)
    ObjectiveFunction *goal;
    /// The dynamic variable ordering heuristic
    DVO  *heuristic;
    /// The different weight and quantities "learned" during search
    Vector<Weighter*> learners;

    /// The current status (UNSAT/SAT/LIMITOUT)
    int         status;
    /// The current level in the search tree (number of left branches from the root)
    int          level;
    int     init_level;
    /// The target backtrack level 
    int backtrackLevel;
    /// The number of search variables
    int         length;
    /// The number of variables
    int        numvars;

    /// The instantiation sequence, an array containing all variables, ordered until the index level-1
    VariableInt   **sequence;
    /// A pointer to the first unassigned variable in sequence
    VariableInt     **future;
    /// A pointer to the end of the array 'sequence'
    VariableInt      **empty;
    /// A pointer to the variables that do not participate in the search
    VariableInt **auxilliary;

  
    /// The array of variables for which a decision has been taken
    Vector< VariableInt* > decision;
    /// An array to store the mins for last solution found
    int     *solution;
    int     *max_solution;

    /// The list of Variables
    Vector < VariableInt* > variables;
    /// The list of Constraints
    Vector < Constraint* > constraints;
    /// The list of unary Constraints
    Vector < UnaryConstraint* > unaryCons;
    /// The list of constants used in the model
    std::map<int, VariableInt*> constants;


    /******************************************
     * Backtracking structures
     ******************************************/
  
  private:

    ///// Pool of Boolean domains
    //ReversibleSparseSet domainPool;
    /// Trail of reversible objects
    Vector<ReversibleObj*> store;
    /// Trail of levels
    Vector<int> lvl_;
    /// Trail of unlinked search variables
    Vector<VariableInt **> past;
    /// Trail of unlinked non-search variables
    Vector<VariableInt **> auxv;  

  public:


    /// GAC Stack
    MistralGacList<VariableInt*> gacvarstack;
    /// GAC Stack
    MistralGacStack<Constraint*> gacconstack;
    /// List of unary constraints

    /// Number of solutions to find, if equal to -1, then all solutions are listed
    int FIND_ALL;
    /// Used for lds search
    int DISCREPANCY;
    /// Number of nodes, that is recursive calls to  the dfs algo
    unsigned long int NODES; 
    /// Limit on the number of nodes
    unsigned long int NDSLIMIT;
    /// Number of backtracks, that is unsuccesful recursive calls 
    unsigned long int BACKTRACKS;
    /// Limit on the number of backtracks, used for restarting
    unsigned long int BTSLIMIT;
    /// Number of constraint failures
    unsigned long int FAILURES; 
    /// Limit on the number of failures, used for probing
    unsigned long int FAILLIMIT;
    /// Number of calls to a constraint propagator
    unsigned long int PROPAGS;
    /// Number of constraint checks (unused)
    unsigned long int CHECKS;
    /// Number of solutions found so far
    unsigned long int SOLUTIONS;
    /// Number of solutions found so far
    unsigned long int FILTERINGS;
    /// Number of solutions found so far
    unsigned long int RESTORES;
    /// flag to know if there is a limit
    unsigned int LIMIT;
    /// Number of solutions found so far
    unsigned long int MISC;

    /// Number of backtracks for each restart
    Vector<unsigned long int> BTSLIST;

  
    /// timestamp
    double INITTIME;
    /// timestamp
    double STARTTIME;
    /// timestamp
    double ENDTIME;
    /// timestamp
    double TOTTIME;
    /// timestamp
    double SOLTIME;
    /// timestamp
    double TIMELIMIT;

    /// level of verbosity
    int verbosity;

    /// 
    unsigned long int fail_increment;
    //@}


    ConstraintClauseBase *sat;


    /**@name Constructors*/
    //@{
    Solver(CSP& c, VarArray& x, VariableOrdering& v);
    void init(CSP& c, VarArray& x, VariableOrdering& v);
    Solver(CSP& c, VariableOrdering& v);
    void init(CSP& c, VariableOrdering& v);
    Solver(CSP& c, VarArray& x);
    void init(CSP& c, VarArray& x);
    Solver(CSP& c, BuildObject **x, const int l);
    void init(CSP& c, BuildObject **x, const int l);
    Solver(CSP& c);
    void init(CSP& c);
    Solver();
    void init();

    void prebuild(CSP&);
    void build(CSP&, const int n=NOVAL);
    void initSearch( VarArray& X, const int s=0 );
    void initSearch( BuildObject **x, const int n, const int s );
    void initSearch( const int n=0, 
		     const int s=NOVAL, 
		     const bool r=true );

    void add(VariableOrdering& g);
    virtual ~Solver();
    //@}
  

    /**@name Solving methods*/
    //@{
    /// Returns UNSAT/SAT/LIMITOUT
    /**
    inline int iterative_dfs() 
    {
      VariableInt *lastDecision;
      
      while( status == UNKNOWN ) {
	 if( filtering() ) {
	   if( future == empty ) {
	     solutionFound();
	   } else {
	     newNode();
	   }
	 } else {
	   if( !level ) {
	     status = UNSAT;
	   } else if( limitsExpired() ) {
	     status = LIMITOUT;
	   } else {
	     lastDecision = decision[level];    
	     backtrackTo( backtrackLevel );
	     lastDecision->branch->right();
	   }
	 }
       }
       return status;
    }
    */



    inline void newNode( )
    {
      if(level++) learnSuccess();

      past.push( future );
      auxv.push( auxilliary  );
      lvl_.push( store.size );

      decision.push( heuristic->select() );
      decision[level]->branch->left(); 

      ++NODES;
      int i=learners.size;
      while( i-- )
	learners[i]->notifyChoice( );
    }


    inline int iterative_dfs() 
    {
      VariableInt *lastDecision;
      while( status == UNKNOWN ) {

	//       for(int k=0; k<length; ++k) {
	// 	if(variables[k]->isGround())
	// 	  std::cout << " " << variables[k]->value();
	// 	else
	// 	  std::cout << " *";
	//       }
	//       std::cout << std::endl;

	if( filtering() ) {
	  if( future == empty ) {
	  
#ifdef _DEBUGSEARCH
	    std::cout << "c";
	    for(int k=0; k<=level; ++k) std::cout << " ";	    
	    std::cout << " SAT!" << std::endl; 
#endif
	  
	    solutionFound(init_level);
	  } else {
	    newNode();
	  
#ifdef _DEBUGSEARCH
	    std::cout << "c";
	    for(int k=0; k<=level; ++k) std::cout << " ";
	    std::cout << " ";
	    decision[level]->branch->printLeft( std::cout );
	    std::cout << std::endl;
#endif
	  
	  }
	} else {
	  
	  // #ifdef _DEBUGSEARCH
	  // 	std::cout << "c";
	  // 	for(int k=0; k<=level; ++k) std::cout << " ";	    
	  // 	std::cout << " failure at level " << level << std::endl; 
	  // #endif
	  if( level <= init_level ) {
	  
#ifdef _DEBUGSEARCH
	    std::cout << "c UNSAT!" << std::endl;
#endif
	  
	    status = UNSAT;
	  } else if( limitsExpired() ) {
	  
#ifdef _DEBUGSEARCH
	    std::cout << "c";
	    for(int k=0; k<=level; ++k) std::cout << " ";	    
	    std::cout << " limit expired at level " << level << std::endl; 
#endif
	  
	    status = LIMITOUT;
	  } else {
	    lastDecision = decision[level];    
	  
#ifdef _DEBUGSEARCH
	    if( level > backtrackLevel+1 ) {
	      std::cout << "c";
	      for(int k=0; k<=level; ++k) std::cout << " ";
	      std::cout << " backjump to level " << backtrackLevel << std::endl;
	    }
#endif

	    backtrackTo( backtrackLevel );

#ifdef _DEBUGSEARCH
	    std::cout << "c";
	    for(int k=0; k<=level; ++k) std::cout << " ";
	    std::cout << " ";
	    lastDecision->branch->printRight( std::cout );
	    std::cout << std::endl;
#endif
	  
	    lastDecision->branch->right();
	  }
	}
      }
      return status;
    }
  
    /*!
      Decomposed dfs
    */
    int startNewSearch();
    /*!
      Decomposed dfs
    */
    int getNextSolution();
    /*!
      Preprocessing step (closure computed with filtering())
    */
    bool preprocess();
  
    /// Returns UNSAT/SAT/LIMITOUT
    /*!
      presolving method
    */
    int presolve();
    /*!
      Simple solving method
    */
    int solve();
    /*!
      LDS solving method
    */
    int ldSolve( const int step=1, const int limit=NOVAL );
    int ldSolve( const int*, const int step=1, const int limit=NOVAL );
    int ldSolve( const int, BuildObject**, int*, const int step=1, const int limit=NOVAL );
    int ldSolve( VarArray&, int*, const int step=1, const int limit=NOVAL );
    int ldSolve( ConstraintLDS*, const int step=1, const int limit=NOVAL );
    bool restart(const double decay=0.0);
    /// Returns UNSAT/SAT/LIMITOUT
    /*!
      Solve using restarts. There is no randomization besides
      that induced by the variables and values ordering.
    */
    int solve_and_restart( const int policy = GEOMETRIC, 
			   const unsigned int base = 32, 
			   const double factor = 1.3333333,
			   const double decay = 0.0 ); 
    /*!
      Diarmuid's random probes
    */
    int random_probe( const unsigned int iterations = 100, 
		      const unsigned int limit = 30 );
    /*!
      singleton arc consistency
    */
    int sacPreprocess( const bool complete = false, 
		       const int wtype = (Weighter::IPT | Weighter::WLD) );
    /// GAC closure
    /*!
      Implementation of the AC3 algorithm
      return the level to which we should backtrack
    */
    bool filtering();
    /*!
      Called when a solution is found. 
    */
    int solutionFound(int init_level);
    /*!
      Check the solution for consistency
    */
    int checkSolution();
    /*!
      Scale down the weights
    */
    inline int weightDecay( double decay)
    {
      int old_weight, new_weight, i=constraints.size, j;
      int threshold = (int)ceil(1.0 / (1.0 - decay));

      while( i-- )
	{
	  old_weight = constraints[i]->weight;
	  if( old_weight >= threshold )
	    {
	      new_weight = (int)ceil(decay * (double)old_weight);
	      j = constraints[i]->arity;
	      while( j-- )
		constraints[i]->scope[j]->weight += (new_weight - old_weight); 
	    }
	}

      return 1;
    }
    /*! 
      Call the individual restore() methods of Variable
      and Constraint
    */
    inline void restore()
    {
      decision.pop();
      VariableInt **vp, **va;
      int lvl;

      past.pop( vp );
      auxv.pop( va );
      lvl_.pop( lvl );

      while( store.size > lvl ) {
	store.back()->restore();
	store.pop();
      }

      while(vp != future) {
	future[-1]->link( );
      }
    
      while(va != auxilliary) 
	auxilliary[-1]->link( );      
    }

    /// Undo search
    inline void backtrackTo( const int lvl )
    {
      while( level > lvl ) {
	++BACKTRACKS;
	restore();
	--level;
      }
    }

    /// Triggers an event e for the variable x
    inline void triggerEvent(const int x, const int e) 
    {
      gacvarstack.push( x, e );
    }

    /// Check if one limit has expired
    inline bool limitsExpired() 
    {
    
#ifdef _DEBUGSEARCH
      if(TIMELIMIT > .0 && getRunTime() - STARTTIME >= TIMELIMIT)
	std::cout << "TIME LIMIT " << (getRunTime() - STARTTIME) << std::endl;
    
      if(BTSLIMIT  >  0 && BACKTRACKS   >=  BTSLIMIT)    
	std::cout << "BTS LIMIT " << BACKTRACKS << " " << BTSLIMIT << std::endl;
    
      if(NDSLIMIT  >  0 && NODES   >=  NDSLIMIT)    
	std::cout << "NDS LIMIT " << NODES << " " << NDSLIMIT << std::endl;
    
      if(FAILLIMIT == 1 && FAILURES   >=  FAILLIMIT)   
	std::cout << "FAIL LIMIT " << FAILURES << " " << FAILLIMIT << std::endl;
#endif
    
      return (
	      (TIMELIMIT > .0 && getRunTime() - STARTTIME >= TIMELIMIT) ||
	      (BTSLIMIT  >  0 && BACKTRACKS   >=  BTSLIMIT)      ||
	      (NDSLIMIT  >  0 && NODES   >=  NDSLIMIT)      ||
	      (FAILLIMIT  >  0 && FAILURES   >=  FAILLIMIT)      //DG Change
	      );
    }
    //@}

    /**@name Helpers*/
    //@{
    void binds( ReversibleObj& x )
    {
      x.level = &level;
      x.store = &store;
    }

    inline void stopWatch()
    {
      ENDTIME = getRunTime() - STARTTIME;
      TOTTIME = (ENDTIME + STARTTIME - INITTIME);
    }

    inline void learnFailure( Constraint *con )
    {
      ++FAILURES;
      int i=learners.size;
      while( i-- )
	learners[i]->notifyFailure( con );
    }

    inline void learnSuccess( )
    {
      int i=learners.size;
      while( i-- ) {
	learners[i]->notifySuccess( );
      }
    }


    //void setHeuristic(string hname, const int rdz);
    void setHeuristic(const char* var_name, const char* val_name, const int rdz);
    void setBranching(BuildObject **x, const int l, const char* val_name);

    void setRandomValueOrdering();
    void setGuidedOrdering(VarArray&, int*);
    void setRandGuidedOrdering(VarArray&, int*, int*, int*);

    /// Cutoff in number of backtracks     
    void setNodeLimit     (unsigned long nl){ NDSLIMIT  = nl; LIMIT = 1; }
    /// Cutoff in number of backtracks     
    void setBacktrackLimit(unsigned long nl){ BTSLIMIT  = nl; LIMIT = 1; }
    /// Cutoff in number of failures  
    void setFailureLimit  (unsigned long nl){ FAILLIMIT = nl; LIMIT = 1; }
    /// Cutoff in cpu time (s)
    void setTimeLimit     (double tl)       { TIMELIMIT = tl; LIMIT = 1; }  
    /// Print satisfiability, number of backtracks and cpu time
    void setVerbosity     (const int v)     { verbosity =  v; }  
    /// Add a learner
    Weighter* setLearner( int wtype );
    /// Use domain splitting
    void setDomainSplitting() { domainSplitting = true; }
    /// Shuffle the array of variables
    void randomizeSequence();
    /// Randomize the array of variables before restarts
    void setRandomized( const bool r=true ) { randomizedRestart = r; }
    /// Set the random seed
    void setRandomSeed( const unsigned int seed ) { usrand(seed); }
    /// Setup the nogoods on restart
    void setRestartNogood();
    /// Setup the forget utility for the clause base
    void setForgetfulness( const double f );
    /// Print statistics corresponding to the bitset stats
    void printStatistics(std::ostream &log=std::cout, const int stats=ALL);
    unsigned long int getBacktracks() const ;
    unsigned long int getNodes     () const ;
    unsigned long int getFailures  () const ;
    unsigned long int getChecks    () const ;
    unsigned long int getPropags   () const ;
    double            getTime      () const ;

    void printSequence()
    {
      int i=0;
      std::cout << "[ ";    
      for(; i<length; ++i) {
	if(sequence+i == future)
	  std::cout << " | ";
	else if(sequence+i == empty)
	  std::cout << " ] ";
	sequence[i]->print(std::cout);
	std::cout << " ";
      }
      std::cout << std::endl;
    }
    void print(std::ostream& o) const;
    void printPython() const;
    void printXML(std::ostream& o) const;
    void printWeightProfile(std::ostream& o, int limit, int t) const;
    //@}

  };

};


#endif // _SOLVER_H
