

/*
  Mistral is a constraint satisfaction and optimisation library
  Copyright (C) 2003-2005  Emmanuel Hebrard
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

  The author can be contacted electronically at 
  ehebrard@cse.unsw.edu.au.
*/

#include <cstring>
#include <signal.h>

#include <mistral_sol.h>
#include <mistral_dvo.h>

using namespace Mistral;
using namespace std;

/**********************************************
I/   compute length
II/  build the model
III/ create the array 'sequence'
**********************************************/


// prebuild :-> basic initialisation, invoque prebuild on the model

// build :->  n is an upperbound on level max, invoque build 
// on the model and initialise sequence

// initSearch( X ) :-> put the searchable variables in X in front
// of the array variables and returns the number of searched variables.

// initsearch( n, type ) :-> put all variables of the given type
// in front of the array 'variables+n' and the rest at the back.
// return the number of variables whose size is under a given limit


double RPTIMELIMIT=60.0;

Solver* mistral_solver;
static void Mistral_SIGINT_handler(int signum) {
  cout <<"\nc ****************** INTERRUPTED ******************\n";
  mistral_solver->stopWatch();
  mistral_solver->printStatistics();
  cout <<"c ****************** INTERRUPTED ******************\n";
  exit(1);
  //throw abort_search();
}

/*********************************************
 * Solver
 **********************************************/

void Solver::randomizeSequence()
{

  //cout << "c shuffle the variables" << endl ;

  int i, j, n=(empty-future);
  VariableInt *aux;
  for(i=0; i<n-1; ++i)
    {
      j = i + (randint(n-i));
      aux = future[j];
      future[j] = future[i];
      future[i] = aux;
      future[i]->seqIdx = &future[i];
      future[j]->seqIdx = &future[j];
    }

//   for(i=0; i<n; ++i) {
//     future[i]->printshort( cout );
//     cout << " ";
//   }
//  cout << endl << endl;
}

Weighter* Solver::setLearner( int wtype )
  {

    int i=learners.size;
    if( wtype & Weighter::WLD )
      while( i-- )
	if( learners[i]->getType() == Weighter::WDG ) {
	  delete learners[i];
	  learners.erase( i );
	}

    i=learners.size;
    if( wtype & Weighter::WDG )
      while( i-- )
	if( learners[i]->getType() == Weighter::WLD ) {
	  delete learners[i];
	  learners.erase( i );
	}
    
    i=learners.size;
    while( i-- )
      if( learners[i]->getType() & wtype ) 
	return learners[i];
    if( (wtype & Weighter::WLD) == Weighter::WLD ) {
      learners.push( new WeighterLevelDegree( this ) );
    }
    else if( (wtype & Weighter::WDG) == Weighter::WDG ) {
      learners.push( new WeighterDegree( this ) );
    }
    if( wtype & Weighter::IPT) {
      learners.push( new WeighterImpact( this ) );
    }
    if( wtype & Weighter::SAC) {
      learners.push( new WeighterSAC( this ) );
    }
    if( wtype & Weighter::ISAC) {
      learners.push( new WeighterISAC( this ) );
    }
    if( wtype & Weighter::RNGD) {
      learners.push( new WeighterRestartNogood( this ) );
    }

    return learners.back();
  }


void Solver::init()
{
  usrand( 11041979 );
  randomizedRestart=false;
  boundOptimised=true;
  domainSplitting=false;
  verbosity = 0;
  fail_increment = 0;
  status = UNKNOWN;
  FIND_ALL = 0;
  DISCREPANCY = -1;
  SOLUTIONS = NODES = BACKTRACKS = CHECKS = FAILURES = PROPAGS = FILTERINGS = RESTORES = MISC = 0; 
  INITTIME = getRunTime();
  ENDTIME = TOTTIME = SOLTIME = 0;


  BTSLIMIT  =  0;
  NDSLIMIT  =  0;
  TIMELIMIT = .0;
  FAILLIMIT =  0;
  LIMIT = 0;

  level = -1;
  init_level = 0;

  solution = NULL;
  max_solution = NULL;
  sequence = NULL;
  heuristic = NULL;
  goal = NULL;
  function = NULL;
  sat = NULL;
}

void Solver::prebuild(CSP& c)
{
  if( !c.preBuild() ) {
    status = UNSAT;
  }
}

void Solver::build(CSP& c, const int n)
{
  length = std::min(n, c.numVars());

  c.build( this );
  numvars    =  variables.size;
  sequence = new VariableInt*[numvars];
  decision.init(0,numvars+1);
}

void Solver::initSearch( BuildObject **x, const int l, const int maxsize ) 
{
  int i=0, j=0, n=l, idx;

  // remove from x the variables that are "non-searchable"
  VariableInt *temp;
  BuildObject *bvar;
  while( i<n ) {
    bvar = x[i];

//     bvar->print( cout );
//     cout << " / ";
//     if( bvar->getVariable() )
//       bvar->getVariable()->print( cout );
//     cout << " ";
//     cout << bvar;
//     cout << endl;

    if( bvar->isSearchable() &&
	((temp = bvar->getVariable()) != NULL) ) {
      
      idx = temp->id;
      if( idx != j ) {
	variables[idx] = variables[j];
	variables[j] = temp;
   
	variables[j]->id = j;
	variables[idx]->id = idx;
      }

      ++j;
    } 
    ++i;
  }
  n=j;

//   for(i=0; i<j; ++i)
//     {
//       variables[i]->print( cout );
//       cout << endl;
//     }

//   //  exit(0);

  initSearch( n, maxsize, false );
}



void Solver::initSearch( VarArray& X, const int maxsize ) 
{
  int i=0, j=0, n=X.size(), idx;

  // remove from x the variables that are "non-searchable"
  VariableInt *temp;
  BuildObject *bvar;
  while( i<n ) {
#ifdef _NARRAY
    bvar = X.array_[i];
#else
    bvar = X.array_[i].var_ptr_;
#endif
    if( bvar->isSearchable() &&
	((temp = bvar->getVariable()) != NULL) ) {

     
//        bvar->print( cout );
//        cout << " / ";
//        bvar->getVariable()->print( cout );
//        cout << " ";
//        cout << bvar;
//        cout << endl;
      
      idx = temp->id;
      if( idx != j ) {
	variables[idx] = variables[j];
	variables[j] = temp;
   
	variables[j]->id = j;
	variables[idx]->id = idx;
      }

      ++j;
    } 
    ++i;
  }
  n=j;

//   for(i=0; i<j; ++i)
//     {
//       variables[i]->print( cout );
//       cout << endl;
//     }

//   //  exit(0);

  initSearch( n, maxsize, false );
}

void Solver::initSearch( const int n, 
			 const int maxsize,
			 const bool range ) 
{
  int i, searchable = n, //= ( maxsize < NOVAL ? n : numvars ), 
    end = numvars;


  //cout << maxsize << " " << searchable << endl;
   //   exit(0);


  if( maxsize ) {
    VariableInt* aux;

    //cout << searchable << " " << end << endl;
    //exit(0);

    while( searchable<end )
      {
	aux = variables[searchable];
	if( aux->isGround() || aux->domsize() > maxsize || aux->getType() == VariableInt::VIRTUAL ||
	    (!range && aux->getType() == VariableInt::RANGE) )
	  {
	    variables[searchable] = variables[--end];
	    variables[searchable]->id = searchable;
	    variables[end] = aux;
	    aux->id = end;
	  }
	else {
	  ++searchable;
	}
      }
  } 

  // then copy the rest of variables into sequence
  std::memcpy(sequence, variables.stack_, numvars*sizeof(VariableInt*));


  empty = (sequence+searchable);
  future = sequence;
  auxilliary = empty;


//   cout << searchable << endl;
//   for(VariableInt **iter=future; iter<empty; ++iter) {
//     (*iter)->print( cout );
//     cout << endl;
//   }
//   cout << future << " " << empty << endl;
//   exit(0);



  length   = (empty-sequence);


  gacvarstack.initList(numvars, variables.stack_);
  gacconstack.initStack(constraints.size);


  solution     = new int[numvars];
  max_solution = new int[numvars];
  std::fill( solution,     solution    +numvars, NOVAL );
  std::fill( max_solution, max_solution+numvars, NOVAL );


   decision.push( NULL );
   lvl_.push(0);
   past.push( future );
   auxv.push( auxilliary );


  for(i=0; i<length; ++i) {
    sequence[i]->seqBeg = &future;
    sequence[i]->seqIdx = sequence+i;
  }

  for(i=0; i<(numvars-length); ++i) {
    auxilliary[i]->seqBeg = &auxilliary;
    auxilliary[i]->seqIdx = auxilliary+i;
  }
}
  
Solver::Solver(CSP& c, VarArray& x, VariableOrdering& v) 
{
  init();
  init(c, x, v);
}

void Solver::init(CSP& c, VarArray& x, VariableOrdering& v) 
{
  prebuild( c );
  if( status == UNKNOWN ) { 
    build( c, x.size() );
    initSearch( x );
    heuristic = v.extract( this );
  }
}

Solver::Solver(CSP& c, VarArray& x) 
{
  init();
  init(c, x);
}

void Solver::init(CSP& c, VarArray& x) 
{
  prebuild( c );
  if( status == UNKNOWN ) { 
    build( c, x.size() );
    initSearch( x );
  }
}

Solver::Solver(CSP& c, BuildObject **x, const int l) 
{
  init();
  init(c, x, l);
}

void Solver::init(CSP& c, BuildObject **x, const int l) 
{
  prebuild( c );
  if( status == UNKNOWN ) { 
    build( c, l );
    initSearch( x, l, 0 );
  }
}

Solver::Solver(CSP& c, VariableOrdering& v) 
{
  init();
  init(c, v);
}

void Solver::init(CSP& c, VariableOrdering& v) 
{
  prebuild( c );
  if( status == UNKNOWN ) { 
    build( c );
    initSearch();
    heuristic = v.extract( this );
  }
}

Solver::Solver(CSP& c) 
{
  init();
  init(c);
}

void Solver::init(CSP& c) 
{
  prebuild( c );
  if( status == UNKNOWN ) { 
    build( c );
    initSearch();
  }
}

Solver::Solver() 
{
  init();
}

void Solver::add(VariableOrdering& v)
{
  DVO *old_heuristic = heuristic;
  heuristic = v.extract( this );
  if( old_heuristic ) delete old_heuristic;
}


Solver::~Solver() 
{
  
  //std::cout  << "*** DELETE SOLVER ***" << endl;

  delete heuristic;
  delete goal;
  delete function;

  delete [] solution;
  delete [] max_solution;
  delete [] sequence;
  
  int i;
  for(i=0; i<constraints.size; ++i) 
    delete constraints[i];

  for(i=0; i<variables.size; ++i)
    delete variables[i];


  i=learners.size;
  while( i-- ) 
    {
      delete learners[i];
    }  
}



int Solver::presolve()
{  

#ifdef _DEBUGAC
  cout << "\tPRESOLVE" << endl;
  for(int k=0; k<numvars; ++k) {
    variables[k]->print( cout );
    cout << endl;
  }
  cout << endl;
#endif

  if( level < 0 )
    {
      STARTTIME = getRunTime();

      ++level;

      if( heuristic == NULL )
	//heuristic = new DVONoOrder( this );
	heuristic = new GenericDVO<VarSelectorDomain>( this );
      
      int i = constraints.size;
      while( i-- ) {

#ifdef _DEBUGPROPAG
	cout << i << " PROPAGATE \n" ;
	constraints[i]->print( cout );
	cout << endl;
	for(int a=0; a<constraints[i]->arity; ++a) {
	  constraints[i]->scope[a]->print( cout );
	  cout << " ";
	}
	cout << endl;
#endif
	
	++PROPAGS;
	if( !(constraints[i]->propagate( )) ) {
	  status = UNSAT;
	  return status;
	}

#ifdef _DEBUGPROPAG
	for(int a=0; a<constraints[i]->arity; ++a) {
	  constraints[i]->scope[a]->print( cout );
	  cout << " ";
	}
	cout << endl;
	constraints[i]->print( cout );
	cout << " OK" << endl << endl;
#endif

      }

      i = unaryCons.size;
      while( i-- ) {
	++PROPAGS;
	if( !(unaryCons[i]->propagate( )) ) {
	  status = UNSAT;
	  return status;
	}
      }
      unaryCons.size = 0;
      
      i = numvars;
      while( i-- )
	if( variables[i]->isGround() ) 
	  //triggerEvent( variables[i], Constraint::VALUETRIGGER );
	  triggerEvent( variables[i]->id, Constraint::VALUETRIGGER );


      i = length;
      while( i-- )
	if( !variables[i]->isGround() && !variables[i]->branch ) {
	  if( domainSplitting && variables[i]->domsize() > 3 && variables[i]->getType() != VariableInt::LIST ) {
	    variables[i]->branch = new ValSelectorSplit( variables[i] );
	  } else if( variables[i]->getType() == VariableInt::BOOL ) {
	    //variables[i]->branch = new ValSelectorRandMinMax( variables[i] );
	    variables[i]->branch = new ValSelectorMin( variables[i] );
	  } else if( variables[i]->getType() == VariableInt::RANGE ) {
	    //variables[i]->branch = new ValSelectorRandMinMax( variables[i] );
	    variables[i]->branch = new ValSelectorMin( variables[i] );
	  } else if( variables[i]->getType() == VariableInt::BIT ) {
	    //variables[i]->branch = new ValSelectorRandMinMax( variables[i] );
	    variables[i]->branch = new ValSelectorMin( variables[i] );
	  } else if( variables[i]->getType() == VariableInt::LIST ) {
	    //variables[i]->branch = new ValSelectorFirst( variables[i] );
	    variables[i]->branch = new ValSelectorMin( variables[i] );
	    //variables[i]->branch = new ValSelectorMax( variables[i] );
	    //variables[i]->branch = new ValSelectorRand( variables[i] );
	    //variables[i]->branch = new ValSelectorRandMinMax( variables[i] );
	  } else {
	    variables[i]->branch = new ValSelectorFirst( variables[i] );
	  }
	}
    }

#ifdef _DEBUGAC
  for(int k=0; k<numvars; ++k) {
    variables[k]->print( cout );
    cout << endl;
  }
  cout << endl;
#endif

  return status;
}


int Solver::solve() 
{
  mistral_solver = this;
  signal(SIGINT,Mistral_SIGINT_handler);

  try {
    if( verbosity > 0 ) 
      cout << "c ===============[ Mistral - solve ]===============" << endl;
    if( randomizedRestart ) randomizeSequence();
    if( status == UNKNOWN ) {
      if( presolve() == UNKNOWN ) {
	if(function)
	  function->initialise();
	iterative_dfs( );
      }
      stopWatch();
    }
  } catch (const abort_search& ex) {
    std::cout << "c *        Abort search exception caught!         *" << std::endl
	      << "c *************************************************" << std::endl;
    status = UNKNOWN;
    //exit(1);
  } catch (...) {
    std::cout << "Fatal error, unkown exception" << std::endl;
    exit(1);
  }
  return status;
}

int Solver::ldSolve(VarArray& scp, int *ideal, const int step, const int limit)
{
  BuildObject* scope[scp.size()];
  for(int i=0; i<scp.size(); ++i)
    scope[i] = scp[i].var_ptr_;
  return ldSolve(scp.size(), scope, ideal, step, limit);
}

int Solver::ldSolve(const int n, BuildObject** scope, int *ideal, const int step, const int limit) 
{
  if( status == UNKNOWN ) {
    if( presolve() == UNKNOWN ) {
      ConstraintHamming *lds = new ConstraintHamming( this, n, scope, ideal );
      //constraints.push( lds );
      status = ldSolve( lds, step, limit );
    }
    stopWatch();
  }
  return status;
}

int Solver::ldSolve(const int *ideal, const int step, const int limit) 
{
  if( status == UNKNOWN ) {
    if( presolve() == UNKNOWN ) {
      ConstraintHamming *lds = new ConstraintHamming( this, ideal );
      //constraints.push( lds );
      status = ldSolve( lds, step, limit );
    }
    stopWatch();
  }
  return status;
}
int Solver::ldSolve(const int step, const int limit) 
{
  if( status == UNKNOWN ) {
    if( presolve() == UNKNOWN ) {
      ConstraintDistance *lds = new ConstraintDistance( this );
      //constraints.push( lds );
      //int ub = lds->getMaxThreshold() ;
      status = ldSolve( lds, step, limit );
    }
    stopWatch();
  }
  return status;
}

int Solver::ldSolve(ConstraintLDS *lds, const int step, const int limit) 
{
  mistral_solver = this;
  signal(SIGINT,Mistral_SIGINT_handler);

  try {
    if( status == UNKNOWN ) {
      int ub = std::min(lds->getMaxThreshold(), limit);
      
      status = UNSAT;
      init_level = 1;
      
      ++DISCREPANCY;
      while( status == UNSAT && DISCREPANCY < ub ) {
	
	++level;
	past.push( future );
	auxv.push( auxilliary  );
	lvl_.push( store.size );
	decision.push( NULL );
	
	lds->lb_threshold = DISCREPANCY;    
	lds->ub_threshold = DISCREPANCY+step-1;
	
	if( verbosity > 0 )
	  {
	    ENDTIME = getRunTime() - STARTTIME;
	    cout << "c" << setw(10) << " " << "discrepancy: " 
		 << setw(5) << DISCREPANCY << " to " 
		 << (DISCREPANCY+step-1) << "  ( " << setw(4) << ENDTIME 
		 << " s)" << endl;
	  }
	
	
	//decision.push( NULL );
	//lvl_.push(0);
	//past.push( future );
	//auxv.push( auxilliary );
	
	
	//recursive_dfs();
	
	
#ifdef _DEBUGPROPAG
	cout << " PROPAGATE \n" ;
	lds->print( cout );
	cout << endl;
	for(int a=0; a<lds->arity; ++a) {
	  lds->scope[a]->print( cout );
	  cout << " ";
	}
	cout << endl;
#endif
	
	++PROPAGS;
	bool lds_propagation = lds->propagate();
	
#ifdef _DEBUGPROPAG
	for(int a=0; a<lds->arity; ++a) {
	  lds->scope[a]->print( cout );
	  cout << " ";
	}
	cout << endl;
	lds->print( cout );
	cout << endl;
	if(lds_propagation) 
	  cout << " OK" << endl << endl;
	else
	  cout << " FAIL" << endl << endl;
#endif
      
	if( !lds_propagation || !filtering() ) {
	  status = UNSAT;
	} else {
	  
	  if(function)
	    function->initialise();
	  
	  status = UNKNOWN;
	  iterative_dfs();
	
	}
	
	if( status == UNSAT ) {
	  restore();
	  --level;
	}
	
	DISCREPANCY += step;
      }
    }
  } catch (const abort_search& ex) {
    std::cout << "c *        Abort search exception caught!         *" << std::endl
	      << "c *************************************************" << std::endl;
    status = UNKNOWN;
    //exit(1);
  } catch (...) {
    std::cout << "Fatal error, unkown exception" << std::endl;
    exit(1);
  }
  return status;
}


int binlog( unsigned long int x ) {
  if( !x ) return NOVAL;
  if( x == 1 ) return 0;
  int exponent = 0;
  while( (x >> (++exponent)) > 1 );
  return exponent;
}
unsigned long int luby_seq( unsigned long int iteration ) {
  int thelog = binlog( iteration );
  if( (int)iteration == (1 << (thelog + 1))-1 )
    return (1 << thelog);
  return luby_seq( iteration - (1 << thelog) + 1 );
}

bool Solver::restart(const double decay) {
  if( randomizedRestart ) randomizeSequence();
  int i=learners.size;
  
  if(function)
    function->initialise();
  
  while( i-- )
    learners[i]->notifyRestart( );
  
  if( iterative_dfs() == LIMITOUT ) {
    status = UNKNOWN;
    backtrackTo(0);
    if( decay > 0.0 )
      weightDecay( decay );
    if( sat ) {
      sat->forget();
    }
  } 
  
  if(BTSLIST.empty())
    BTSLIST.push(BACKTRACKS);
  else
    BTSLIST.push(BACKTRACKS - BTSLIST.back());
  
  
  FAILLIMIT = (FAILURES + fail_increment);
  return( (NDSLIMIT <= 0 || NODES < NDSLIMIT) && 
	  (TIMELIMIT <= .0 || (getRunTime() - STARTTIME < TIMELIMIT)) );
}

int Solver::solve_and_restart( const int policy, 
			       const unsigned int base, 
			       const double factor,
			       const double decay)
{
  mistral_solver = this;
  signal(SIGINT,Mistral_SIGINT_handler);

  try {
    if( status == UNKNOWN ) {
      ////////////////////////////////////////////////////////////
      //double PRESOLVETIME = getRunTime();
      ////////////////////////////////////////////////////////////


      int numvar, numval, numcon;

      if( verbosity > 1 )
	cout << endl 
	     << "c  ==================================[ Mistral ]=====================================" << endl
	     << "c  |      SEARCH  STATISTICS           |           PROBLEM STATISTICS               |" << endl
	     << "c  |     Fails  |       Nodes  CPUTime |  #Variables       #values     #constraints |" << endl
	     << "c  ==================================================================================" << endl;
      else if( verbosity > 0 ) 
        cout << "c ==============[ Mistral - restarts ]=============" << endl
	     << "c | restart |    limit |   # of nodes |  cpu time |" << endl
	     << "c =================================================" << endl;


      if( presolve() == UNKNOWN ) {
 
	unsigned int iteration = 0;
	FAILLIMIT = base+FAILURES;
	fail_increment = base; 
	while( status == UNKNOWN ) {
	  ++iteration;
	  switch( policy ) {
	  case GEOMETRIC : fail_increment = 
	      lround((double)(fail_increment * factor));
	    break;
	  case LUBY : fail_increment = 
	      lround((double)(base * luby_seq(iteration)));
	    break;
	  default : fail_increment = 
	      lround((double)(base * luby_seq(iteration) * factor));
	  }
      
      
	  if( verbosity > 1 ) {
	  
	    numvar = (empty - future);
	    numcon = constraints.size;
	    numval = 0;
	    for(int i=0; i<length; ++i) 
	      numval += sequence[i]->domsize();
	  
	    ENDTIME = getRunTime() - STARTTIME;
	    if( ENDTIME < 0.001 ) ENDTIME = 0.001;
	  
	    cout << "c  | " << setw(10) << (FAILLIMIT-FAILURES)
		 << " | " << setw(10) << NODES
		 << " " << setw(9) << ENDTIME
		 << " | " << setw(11) << numvar
		 << "  " << setw(13) << numval
		 << "  " << setw(13) << numcon
		 << "  |" << endl;
	  } else if( verbosity > 0 ) {

	    ENDTIME = getRunTime() - STARTTIME;
	    if( ENDTIME < 0.001 ) ENDTIME = 0.001;
	 

	    cout << "c | " << left  << setw(7)  << (iteration) 
		 << " | "  << right << setw(8)  << (FAILLIMIT-FAILURES)
		 << " | "  << right << setw(12) << NODES     
		 << " | "  << right << setw(9) << ENDTIME 
		 << " |"   << endl;


	  }

// 	  if( randomizedRestart ) randomizeSequence();
// 	  int i=learners.size;

// 	  if(function)
// 	    function->initialise();

// 	  while( i-- )
// 	    learners[i]->notifyRestart( );
	
// 	  if( iterative_dfs() == LIMITOUT ) {
// 	    status = UNKNOWN;
// 	    backtrackTo(0);
// 	    if( decay > 0.0 )
// 	      weightDecay( decay );
// 	    if( sat ) {
// 	      sat->forget();
// 	    }
// 	  } 

// 	  //if( sat )
// 	  //BTSLIST.push( sat->stats.learnt_avg_size );
// 	  if(BTSLIST.empty())
// 	    BTSLIST.push(BACKTRACKS);
// 	  else
// 	    BTSLIST.push(BACKTRACKS - BTSLIST.back());
	    

// 	  FAILLIMIT = (FAILURES + fail_increment);
// 	  if( (NDSLIMIT > 0 && NODES >= NDSLIMIT) || 
// 	      (TIMELIMIT > .0 && getRunTime() - STARTTIME >= TIMELIMIT) ) break;

	  if(!restart(decay)) break;

	}
      } else {

	cout << "STATUS " << status << endl;
      }
      stopWatch();
    } 


    if( verbosity > 1 ) {
      cout << "c  ==================================================================================" << endl;
      if( status == SAT ) 
	cout << "c  |                               SATISFIABLE!                                     |" << endl;
      else if( status == UNSAT ) 
	cout << "c  |                              UNSATISFIABLE!                                    |" << endl;
      else if( status == UNKNOWN ) 
	cout << "c  |                                 UNKNOWN!                                       |" << endl;
      if( ENDTIME < 0.001 ) ENDTIME = 0.001;
      cout << "c  ==================================================================================" << endl ;
      cout << "c  |      Fails  |      Nodes  |       Propags  |  CPU Time |   Fails/s |   Nodes/s |" << endl
	   << "c  | " << setw(10) << FAILURES
	   << "  | " << setw(10) << NODES
	   << "  | " << setw(13) << PROPAGS
	   << "  | " << setw(9) << ENDTIME
	   << " | " << setw(9) << int(double(FAILURES)/ENDTIME) 
	   << " | " << setw(9) << int(double(NODES)/ENDTIME) 
	   << " |" << endl
	   << "c  ==================================================================================" << endl << endl;
    } else if( verbosity > 0 ) 
      cout << "c =================================================" << endl;

  } catch (const abort_search& ex) {
    std::cout << "c *        Abort search exception caught!         *" << std::endl
	      << "c *************************************************" << std::endl;
    status = UNKNOWN;
    //exit(1);
  } catch (...) {
    std::cout << "Fatal error, unkown exception" << std::endl;
    exit(1);
  }
  
  return status;
}



int Solver::random_probe( const unsigned int iterations, 
			  const unsigned int limit )
{
  mistral_solver = this;
  signal(SIGINT,Mistral_SIGINT_handler);

  if( status == UNKNOWN ) {
    if( presolve() == UNKNOWN ) {

      int i = iterations;
      status = LIMITOUT;
    
      while( i-- && status == LIMITOUT ) {
	status = UNKNOWN;

	if( verbosity > 0 && !((i+1) % (iterations/10)))
	  cout << ".";
      
	FAILLIMIT = FAILURES+limit;

	if( randomizedRestart ) randomizeSequence();      
	iterative_dfs();
	if( status == LIMITOUT )
	  backtrackTo(0);

	if( (NDSLIMIT > 0 && NODES >= NDSLIMIT) 
	    || ((getRunTime() - STARTTIME) >= RPTIMELIMIT) 
	    ) 
	  break;
      }
   
      FAILLIMIT = 0;
      
      if( status == LIMITOUT )
	status = UNKNOWN;
    } 

    if(status != UNKNOWN)
      stopWatch();
  }
  
  return status;
}


int Solver::sacPreprocess( const bool comp, const int wtype )
{
  mistral_solver = this;
  signal(SIGINT,Mistral_SIGINT_handler);

  WeighterSAC *ws = NULL;
  WeighterImpact *wi = NULL;
  
  if( presolve() == UNKNOWN ) {

    if( (wtype & Weighter::WLD) == Weighter::WLD )
      setLearner( Weighter::WLD );
    else if( wtype & Weighter::WDG )
      setLearner( Weighter::WDG );
    if( wtype & Weighter::IPT )
      wi = ((WeighterImpact*)(setLearner( Weighter::IPT )));
    
    if( wtype & Weighter::IPT ){
      ws = (WeighterISAC*)(setLearner( Weighter::ISAC )); // DG Added for doing sac initialization of impacts
      ws->complete = comp;
      
      
      for( int i=0; i<length; ++i ){
	if( variables[i]->getType() != VariableInt::RANGE )
	  std::fill( wi->decision_impact[i]+variables[i]->minCapacity(), wi->decision_impact[i]+variables[i]->maxCapacity(), 0 );
	else 
	  wi->decision_impact[i][0] = wi->decision_impact[i][1] = 0;
      }
    }
    else
      {  ws = (WeighterSAC*)(setLearner( Weighter::SAC ));
	ws->complete = comp;
	cout << "SAC if statement" << endl;
      }
    
    DVOSingletonAC *h = new DVOSingletonAC( this, ws );
    
    DVO *oldh = heuristic;
    heuristic = h;
    
    
    ValSelector* oldvo[length];
    for( int i=0; i<length; ++i )
      {
	oldvo[i] = variables[i]->branch;
	if( h->isRange[i] )
	  variables[i]->branch = new ValSelectorSacRange( variables[i], &(h->SACdomain[i]) );
	else
	  variables[i]->branch = new ValSelectorSac( variables[i], &(h->SACdomain[i]) );
      }


    status = LIMITOUT;    
    while( status == LIMITOUT && ws->isNotSac() )
      {
	status = UNKNOWN;
      
	FAILLIMIT = FAILURES+1;

	iterative_dfs();

	//cout << endl << (status == SAT ? "SAT" : (status == UNSAT ? "UNSAT" : (status == LIMITOUT ? "LIMITOUT" : "UNKNOWN"))) << endl;

	if( status == LIMITOUT )
	  backtrackTo(0);
      }

    
    FAILLIMIT = 0;


    delete h;			
    heuristic = NULL;

    if( oldh )
      heuristic = oldh;

    for( int i=0; i<length; ++i )
      {
	delete variables[i]->branch;
	variables[i]->branch = oldvo[i];
      }
    
    delete learners.back();
    learners.pop();
  }

  if( status == LIMITOUT ) status = UNKNOWN;

  return status;
}


int Solver::checkSolution() 
{
  int max_arity = 0;
  int i, j, *aux;
  Vector<int> inconsistencies;
  bool assigned;

  for(i =0; i<constraints.size; ++i) {
    if(max_arity < constraints[i]->arity)
      max_arity = constraints[i]->arity;
  }

  int sol[max_arity];

  for(i =0; i<constraints.size; ++i) {
    j = constraints[i]->arity;

 //    while(j--)
//       sol[j] = constraints[i]->scope[j]->min();

//     if( constraints[i]->check(sol) ) {
//       inconsistencies.push(i);

//       exit(0);
//       }

    assigned = true;
    while(j--) {
      if(!(constraints[i]->scope[j]->isGround())) assigned = false;
      sol[j] = constraints[i]->scope[j]->first();
    }

    if(assigned) {
      if( constraints[i]->check(sol) )
	inconsistencies.push( i );
    } else {
      aux = constraints[i]->supp;
      constraints[i]->supp = sol;
      if( !constraints[i]->findSupport(0, sol[0]) ) 
      inconsistencies.push( i );
      constraints[i]->supp = aux;
    }
  }

  while( inconsistencies.size ) {
    inconsistencies.pop( j );
    cout << "c Error: Solution does not verify\nc\t";
    constraints[j]->print( cout );
    for(i=0; i<constraints[j]->arity; ++i) {
      cout << endl << "c\t";
      constraints[j]->scope[i]->print( cout );
    }
    cout << endl << "s NOT SUPPORTED" << endl << endl;
    exit( 0 );
  }

  return 1;
}

  unsigned long int Solver::getBacktracks() const {return BACKTRACKS;}
  unsigned long int Solver::getNodes     () const {return NODES     ;}
  unsigned long int Solver::getFailures  () const {return FAILURES  ;}
  unsigned long int Solver::getChecks    () const {return CHECKS    ;}
  unsigned long int Solver::getPropags   () const {return PROPAGS   ;}
  double            Solver::getTime      () const {return TOTTIME   ;}


void Solver::printStatistics(std::ostream &log, const int stats)
{

  if(ENDTIME < .001)
    ENDTIME = .0;

  if(TOTTIME < .001)
    TOTTIME = .0;

  if( stats & OUTCOME ) {
    if(goal) {
      log << std::left << std::setw(30) << "c Best objective" << ":"
	  << std::right << std::setw(20) << (goal->upper_bound) ;
      if(status == UNSAT)
	log << " (optimal)" ;
      else
	log << " (may not be optimal)" ;
      log << std::endl;
    } else if(FIND_ALL || SOLUTIONS > 1) { 
      log << std::left << std::setw(30) << "c #Solutions" << ":"
	  << std::right << std::setw(20) << SOLUTIONS ;
      if(status == UNSAT)
	log << " (complete)" ;
      else
	log << " (may not be complete)" ;
      log << std::endl;
    } else 
      log << std::left << std::setw(30) << "c Result" << ":"
	  << std::right << std::setw(20) << ((status == SAT) ? "SATISFIABLE" : (status == UNSAT ? "UNSATISFIABLE" : "UNKNOWN")) << std::endl;
  }

  if( BTSLIST.empty() )
    BTSLIST.push( BACKTRACKS );

  if( stats & BTS )
    log << std::left << std::setw(30) << "c #Backtracks" << ":"
	<< std::right << std::setw(20) << BACKTRACKS << std:: endl;
  if( stats & NDS )
    log << std::left << std::setw(30) << "c #Nodes" << ":"
	<< std::right << std::setw(20) << NODES << std::endl;
  if( stats & FLS )
    log << std::left << std::setw(30) << "c #Failures" << ":"
	<< std::right << std::setw(20) << FAILURES << std::endl;
  if( stats & PPGS )
    log << std::left << std::setw(30) << "c #Propagations" << ":"
	<< std::right << std::setw(20) << PROPAGS << std::endl;
  if( stats & CKS )
    log << std::left << std::setw(30) << "c #Constraint checks" << ":"
	<< std::right << std::setw(20) << CHECKS << std::endl;
  if( stats & REST )
    log << std::left << std::setw(30) << "c #Restarts" << ":"
	<< std::right << std::setw(20) << BTSLIST.size << std::endl;
  if( stats & LASTBTS ) 
    log << std::left << std::setw(30) << "c #Backtracks on final restart" << ":"
	<< std::right << std::setw(20) << BTSLIST.back() << std::endl; 
  if( stats & RUNTIME )
    log << std::left << std::setw(30) << "c Solving time (in sec.)" << ":"
	<< std::right << std::setw(20) << std::setprecision(5) << ENDTIME << std::endl ;
  if( stats & OPTTIME )
    log << std::left << std::setw(30) << "c Optimisation time (in sec.)" << ":"
	<< std::right << std::setw(20) << std::setprecision(5) << SOLTIME << std::endl ; 
  if( stats & PROOFTIME )
    log << std::left << std::setw(30) << "c Proof time (in sec.)" << ":"
	<< std::right << std::setw(20) << std::setprecision(5) << (ENDTIME-SOLTIME) << std::endl ;  
  if( stats & TOTALTIME )
    log << std::left << std::setw(30) << "c Total time (incl. build)" << ":"
	<< std::right << std::setw(20) << std::setprecision(5) << TOTTIME << std::endl ;
  if( stats & SPEEDB ) {
    log << std::left << std::setw(30) << "c Nodes/second" << ":";
    if(ENDTIME)
      log << std::right << std::setw(20) << (unsigned int)(((double)NODES)/ENDTIME) << std::endl;
    else
      log << std::right << "N/A" << std::endl;
  }
  if( stats & SPEEDP ) {
    log << std::left << std::setw(30) << "c Propagations/second" << ":";
    if(ENDTIME)
      log << std::right << std::setw(20) << (unsigned int)(((double)PROPAGS)/ENDTIME) << std::endl;
    else
      log << std::right << "N/A" << std::endl;
  } 
  log.flush();
}

void Solver::printWeightProfile(std::ostream& o, int limit, int threshold) const
{
  int i, j, aux, orderedvar[length], total=0;
  for(i=0; i<length; ++i)
    orderedvar[i] = i;
  for(i=1; i<length; ++i)
    {
      j=i;
      while( j && variables[orderedvar[j]]->weight > variables[orderedvar[--j]]->weight )
	{
	  aux = orderedvar[j];
	  orderedvar[j] = orderedvar[j+1];
	  orderedvar[j+1] = aux;
	}
    }
  if( length < limit )
    limit = length;
  double n=0;
  for(i=0; i<limit; ++i) {    
    if( variables[orderedvar[i]]->weight > (unsigned int)threshold )
      ++n;
    total += (variables[orderedvar[i]]->weight-1);
  }

  cout << "d TOTAL " << total << " WEIGHTEDVARS " << (n/((double)limit)) << "\nc ";
}

bool Solver::preprocess() 
{
  for(int i=0; i<numvars; ++i) 
    gacvarstack.push(variables[i]->id, Constraint::RANGETRIGGER);
  return filtering();
}

int Solver::solutionFound(int init_level) 
{
//   for(int i=0; i<numvars; ++i) {
//     cout << setw(4) << (i+1) << " " << variables[i]->value() << endl;
//   }

  //cout << "store solution" << endl;

  for(int i=0; i<length; ++i) {
    solution[i] = max_solution[i] = variables[i]->value() ;
  }
  for(int i=length; i<numvars; ++i) {
    solution[i] = variables[i]->value() ;
    max_solution[i] = variables[i]->max() ;
  }

  checkSolution();
  ++SOLUTIONS;

  SOLTIME = getRunTime() - STARTTIME;
  if(SOLTIME < .0001)
    SOLTIME = .0;

  status = SAT;  
  if(function) function->execute();

  if(verbosity > 2  || (goal && verbosity > 0)) {

    if(verbosity > 1) {
      if(FAILLIMIT)
	cout << "c  ----------------------------------------------------------------------------------" << endl;
      cout << "c  | " << std::setw(10) << NODES << " NDS"
	   << std::setw(10) << std::setprecision(5)
	   << SOLTIME << " s" ; 
      cout << "            solution: " << std::setw(8) 
	   << (goal ? goal->solution_score() : SOLUTIONS) ;

      if( verbosity > 2 ) {
	cout << ": {" ;
	for(int i=0; i<length-1; ++i) 
	  cout << solution[i] << " ";
	cout << solution[length-1] << "}" ;
      } 
      cout << endl;
      if(FAILLIMIT)
	cout << "c  ----------------------------------------------------------------------------------" << endl;

    } else {
      cout << left  << setw(13) << "c | solution: " 
	   << left  << setw(8)  << (goal ? goal->solution_score() : SOLUTIONS) 
	   << " | " << right << setw(12) << NODES     
	   << " | "  << right << setw(9) << SOLTIME 
	   << " |"   << endl;
    }
  }

  if(goal) {
    if( (status = goal->update()) != SAT ) {
      if( level ) {
	backtrackLevel = level-1;
	VariableInt *lastDecision = decision[level];    
	backtrackTo( backtrackLevel );
	lastDecision->branch->right();
      } else status = UNSAT;
    }

    return status;
  }

  if(FIND_ALL && (int)SOLUTIONS != FIND_ALL) {
    if( level > init_level ) {
      status = UNKNOWN;
      backtrackLevel = level-1;
      VariableInt *lastDecision = decision[level];    
      backtrackTo( backtrackLevel );
      lastDecision->branch->right();
    } else status = UNSAT;
  }

  return status;
}


bool Solver::filtering() 
{
  backtrackLevel = NOVAL;

#ifdef _DEBUGAC
  cout << "\tGAC" << endl;
  for(int k=0; k<numvars; ++k) {
    variables[k]->print( cout );
    cout << endl;
  }
  cout << endl;
#endif

  bool consistent = ( !goal || goal->score() <= goal->upper_bound );
  Constraint *con = NULL;

  if( consistent ) {
    int sz = unaryCons.size, domain_event;
    while( consistent && sz-- ) {
      ++PROPAGS;
      if( !(unaryCons[sz]->propagate( )) ) consistent = false;
    }

    if( consistent ) {
      VariableInt *pvar;
      MistralNode<Constraint*> *nd;

      bool fixedPoint = (gacvarstack.empty() && gacconstack.empty());     
 
      while( consistent && !fixedPoint ) {

	if( !gacvarstack.empty() ) {

#ifdef _DEBUGPROPAG
	  cout << "stack:";
	  int q = gacvarstack.head;
	  do {
	    q = gacvarstack.next[q];
	    cout << " " << q;
	  } while( q != gacvarstack.tail );
	  cout << endl;
#endif

	  pvar = gacvarstack.pop( domain_event );
	  
#ifdef _DEBUGPROPAG
	  if( domain_event & Constraint::VALUETRIGGER ) {
	    cout << "\tValue event on " ;
	    pvar->print( cout );
	    cout << endl;
	    nd = pvar->constraintsOnValue();
	  } else if( domain_event & Constraint::RANGETRIGGER ) {
	    cout << "\tRange event on " ;
	    pvar->print( cout );
	    cout << endl;
	    nd = pvar->constraintsOnRange(); 
	  } else {
	    cout << "\tDomain event on " ;
	    pvar->print( cout );
	    cout << endl;
	    nd = pvar->constraintsOnDomain();
	  }
	  cout << endl;
#endif

	  if( domain_event & Constraint::VALUETRIGGER ) {
	    nd = pvar->constraintsOnValue();
	    pvar->unLink();
	  } else if( domain_event & Constraint::RANGETRIGGER )
	    nd = pvar->constraintsOnRange(); 
	  else 
	    nd = pvar->constraintsOnDomain();
	  
	  while( nextNode(nd) ) {
	    con = nd->elt;
	    if( con->delayed ) {
	      gacconstack.push( con, pvar );
	      continue;
	    }

#ifdef _DEBUGPROPAG
	    cout << "PROPAGATE \n" ;
	    con->print( cout );
	    cout << endl;
	    for(int a=0; a<con->arity; ++a) {
	      con->scope[a]->print( cout );
	      cout << " ";
	    }
	    cout << endl;
#endif

	    ++PROPAGS;
	    if( !(consistent = con->propagate(nd->index, domain_event)) ) {
	      learnFailure( con );
	      break;
	    }

#ifdef _DEBUGPROPAG
	    for(int a=0; a<con->arity; ++a) {
	      con->scope[a]->print( cout );
	      cout << " ";
	    }
	    cout << endl;
	    con->print( cout );
	    cout << " OK" << endl << endl;
#endif
	  }
	  
	} else if(consistent) {
	  if( gacconstack.empty() ) fixedPoint = true;
	  else {
	    con = gacconstack.pop( sz );
	    
#ifdef _DEBUGPROPAG 
	    cout << "DELAYED: " ;
	    variables[sz]->print( cout );
	    cout << endl;
	    for(int a=0; a<con->arity; ++a) {
	      con->scope[a]->print( cout );
	      cout << " ";
	    }
	    cout << endl;
	    con->print( cout );
	    cout << endl;
#endif 	
	    ++PROPAGS;
	    if( !(consistent = con->propagate()) ) {
	      learnFailure( con ); 
	      break;
	    }

#ifdef _DEBUGPROPAG
	    for(int a=0; a<con->arity; ++a) {
	      con->scope[a]->print( cout );
	      cout << " ";
	    }
	    cout << endl;
	    con->print( cout );
	    cout << " OK" << endl << endl;
#endif
	    
	  }
	}
      }
    }
  }

#ifdef _DEBUGPROPAG
  if( !consistent ) {
    if( con )
      con->print( cout );
    cout << " FAILS" << endl << endl;
  }
#endif

  gacvarstack.clear();
  gacconstack.clear();

#ifdef _DEBUGAC
  if( consistent ) {
    for(int k=0; k<numvars; ++k) {
      variables[k]->print( cout );
      cout << endl;
    }
    cout << endl;
  }
  else cout << "FAILURE" << endl;
#endif


  if( !consistent && backtrackLevel > level )
    backtrackLevel = level-1;

  return consistent;
}



int Solver::startNewSearch()
{
  mistral_solver = this;
  signal(SIGINT,Mistral_SIGINT_handler);
  presolve();
  //if( presolve() != UNKNOWN )
  return status;  
}


int Solver::getNextSolution()
{
  if( SOLUTIONS ) {
    if( !level ) {
      status = UNSAT;
      return UNSAT;
    }
    status = UNKNOWN;
    backtrackLevel = level-1;
    VariableInt *lastDecision = decision[level];    
	  
#ifdef _DEBUGSEARCH
    if( level > backtrackLevel+1 ) {
      cout << "c";
      for(int k=0; k<=level; ++k) cout << " ";
      cout << " backjump to level " << backtrackLevel << endl;
    }
#endif
    
    backtrackTo( backtrackLevel );
    
#ifdef _DEBUGSEARCH
    cout << "c";
    for(int k=0; k<=level; ++k) cout << " ";
    cout << " ";
    lastDecision->branch->printRight( cout );
    cout << endl;
#endif
  
    lastDecision->branch->right();
  }

  int res = iterative_dfs();
  stopWatch();
  return res;
}

void Solver::setForgetfulness( const double f )
{
  if( !sat ) {
    sat = new ConstraintClauseBase( this );
    sat->initialise();
  }
  sat->params.forgetfulness = f;
}

void Solver::setRestartNogood()
{
  if( !sat ) {
    sat = new ConstraintClauseBase( this );
    sat->initialise();
  }
  setLearner( Weighter::RNGD );
  ((WeighterRestartNogood*)(learners.back()))->sat = sat;
}

void Solver::printPython() const
{
  print( cout );
}

void Solver::print(std::ostream& o) const
  {
    if(goal) {
      goal->print(o);
      o << endl << "subject to" << endl;
    }

    int i;
    o << "Variables: " << endl;
    for(i = 0; i < variables.size; ++i)
      {
	o << "\t" ;
	variables[i]->print(o);
	o << std::endl;
      }
    o << endl;
    o << "Constraints:" << endl;
    for(i = 0; i < constraints.size; ++i)
      {
	o << "\t" ;
	constraints[i]->print(o);
	o << std::endl;
      }
    o << endl;
  }

void Solver::printXML(std::ostream& o) const
{
  int i,j;

  // header	  
  o << "<?xml version=\"1.0\" encoding=\"UTF-8\"?>" 
    << endl << endl << "<instance>" << endl 
    << "<presentation name=\"?\" format=\"XCSP 2.0\"/>" 
    << endl << endl;
	
  
  // domains
  int nDomains=0;
  bool known;
  BitSet domain[length];
  int domainSize[length];
  int domainIndex[length];
  for(i=0; i<length; ++i) 
    {
      known = false;
      for(j=0; !known && j<nDomains; ++j)
	known = ( sequence[i]->included( domain[j] ) && 
		  sequence[i]->include( domain[j] ) );
      if(!known) {	
	domainIndex[i] = nDomains;
	domain[nDomains].init( sequence[i]->min(), 
			       sequence[i]->max(), 
			       BitSet::empt);
	sequence[i]->unionTo( domain[nDomains] );
	domainSize[i] = sequence[i]->domsize();
	++nDomains;
      } else {
	domainIndex[i] = j;
      }
    }
  o << "<domains nbDomains=\"" << nDomains << "\">" << endl;
  for(i=0; i<nDomains; ++i)
    {
      cout << "<domain name=\"D" << i 
	   << "\" nbValues=\"" << domainSize[i] << "\">" ;
      domain[i].print( cout, "   " );
      cout << "</domain>" << endl; 
    }
  cout << "</domains>" << endl << endl;

}



void Solver::setBranching(BuildObject **x, const int l, const char* val_name)
{
  string Val = val_name;
  VariableInt *temp;
  int i=l;
  if(Val == "Lex") {
    while( i-- ) {
      if( x[i]->isSearchable() && ((temp = x[i]->getVariable()) != NULL) )
	if( !temp->isGround() && !temp->branch ) {
	  if( temp->getType() != VariableInt::INTLIST ) {
	    temp->branch = new ValSelectorMin( temp );
	  } else {
	    temp->branch = new ValSelectorFirst( temp );
	  }
	}
    }
  } else if(Val == "AntiLex") {
    while( i-- )
      if( x[i]->isSearchable() && ((temp = x[i]->getVariable()) != NULL) )
      if( !temp->isGround() && !temp->branch ) {
	if( temp->getType() != VariableInt::INTLIST ) {
	  temp->branch = new ValSelectorMax( temp );
	} else {
	  temp->branch = new ValSelectorFirst( temp );
	}
      }
  } else if(Val == "Random") {
    while( i-- )
      if( x[i]->isSearchable() && ((temp = x[i]->getVariable()) != NULL) )
	if( !temp->isGround() && !temp->branch ) {
	  if( temp->getType() == VariableInt::RANGE ) {
	    temp->branch = new ValSelectorRandMinMax( temp );
	  } else if( temp->getType() != VariableInt::INTLIST ) {
	    temp->branch = new ValSelectorRand( temp );
	  } else {
	    temp->branch = new ValSelectorFirst( temp );
	  }
	}
  } else if(Val == "RandomMinMax") {
    while( i-- )
      if( x[i]->isSearchable() && ((temp = x[i]->getVariable()) != NULL) )
	if( !temp->isGround() && !temp->branch ) {
	  if( temp->getType() != VariableInt::INTLIST ) {
	    temp->branch = new ValSelectorRandMinMax( temp );
	  } else {
	    temp->branch = new ValSelectorFirst( temp );
	  }
	}
  } else if(Val == "DomainSplit") {
    while( i-- )
      if( x[i]->isSearchable() && ((temp = x[i]->getVariable()) != NULL) )
	if( !temp->isGround() && !temp->branch ) {	
	  if( temp->getType() == VariableInt::INTLIST ) {
	    temp->branch = new ValSelectorFirst( temp );
	  } else if( temp->domsize() > 3 && temp->getType() != VariableInt::LIST )
	    temp->branch = new ValSelectorSplit( temp );
	} else {
	  temp->branch = new ValSelectorMin( temp );
	} 
  }
}

//void Solver::setHeuristic(string Heu, const int rdz)
void Solver::setHeuristic(const char* var_name, const char* val_name, const int rdz)
{
  if(status == UNKNOWN) {
    string Heu = var_name;
    string Val = val_name;
    if( rdz > 1 ) {
      if( Heu == "MinDomain" ) 
	heuristic = new GenericRandomDVO<VarSelectorDomain>(this, rdz);
      else if( Heu == "Lex" ) 
	heuristic = new DVOLexicographic(this);
      else if( Heu == "MaxDegree") 
	heuristic = new GenericRandomDVO<VarSelectorDegree>(this, rdz);
      else if( Heu == "MinDomainMinVal") 
	heuristic = new GenericRandomDVO<VarSelectorDomainMinVal>(this, rdz);
      else if( Heu == "Random") 
	heuristic = new DVORandom(this);
      else if( Heu == "MinDomainMaxDegree") 
	heuristic = new GenericRandomDVO<VarSelectorDomainDegree>(this, rdz);
      else if( Heu == "DomainOverDegree") 
	heuristic = new GenericRandomDVO<VarSelectorDomainOverDegree>(this, rdz);
      else if( Heu == "DomainOverWLDegree") {
	setLearner( Weighter::WLD );
	heuristic =  new GenericRandomDVO<VarSelectorDomainOverWeight>(this, rdz);
      }
      else if( Heu == "DomainOverWDegree") {
	setLearner( Weighter::WDG );
	heuristic =  new GenericRandomDVO<VarSelectorDomainOverWeight>(this, rdz);
      }
      else if( Heu == "Neighbour") 
	heuristic =  new GenericRandomDVO<VarSelectorNeighborDD>(this, rdz);
      else if( Heu == "Impact") {
	Impact H(rdz);
	heuristic = H.extract(this);
      }
      else if( Heu == "ImpactOverDegree") {
	ImpactOverDeg H(rdz);
	heuristic = H.extract(this);
      }
      else if( Heu == "ImpactOverWDegree") {
	ImpactOverWDeg H(rdz);
	heuristic = H.extract(this);
      }
      else if( Heu == "ImpactOverWLDegree") {
	ImpactOverWLDeg H(rdz);
	heuristic = H.extract(this);
      }
      else if( Heu == "Scheduling") {
	OSP H(rdz);
	heuristic = H.extract(this);
      }
    } else {
      if( Heu == "No" ) 
	heuristic = new DVONoOrder(this);
      else if( Heu == "MinDomain" ) 
	heuristic = new GenericDVO<VarSelectorDomain>(this);
      else if( Heu == "Lex" ) 
	heuristic = new DVOLexicographic(this);
      else if( Heu == "MaxDegree") 
	heuristic = new GenericDVO<VarSelectorDegree>(this);
      else if( Heu == "MinDomainMinVal") 
	heuristic = new GenericDVO<VarSelectorDomainMinVal>(this);
      else if( Heu == "Random") 
	heuristic = new DVORandom(this);
      else if( Heu == "MinDomainMaxDegree") 
	heuristic = new GenericDVO<VarSelectorDomainDegree>(this);
      else if( Heu == "DomainOverDegree") 
	heuristic = new GenericDVO<VarSelectorDomainOverDegree>(this);
      else if( Heu == "DomainOverWLDegree") {
	setLearner( Weighter::WLD );
	heuristic =  new GenericDVO<VarSelectorDomainOverWeight>(this);
      }
      else if( Heu == "DomainOverWDegree") {
	setLearner( Weighter::WDG );
	heuristic =  new GenericDVO<VarSelectorDomainOverWeight>(this);
      }
      else if( Heu == "Neighbour") 
	heuristic =  new GenericDVO<VarSelectorNeighborDD>(this);
      else if( Heu == "Impact") {
	Impact H;
	heuristic = H.extract(this);
      }
      else if( Heu == "ImpactOverDeg") {
	ImpactOverDeg H;
	heuristic = H.extract(this);
      }
      else if( Heu == "ImpactOverWDeg") {
	ImpactOverWDeg H;
	heuristic = H.extract(this);
      }
      else if( Heu == "ImpactOverWLDeg") {
	ImpactOverWLDeg H;
	heuristic = H.extract(this);
      }
      else if( Heu == "Scheduling") {
	OSP H;
	heuristic = H.extract(this);
      }
    }


    //   cout << endl << Val << endl;

    int i;
    if(Val == "Lex") {

      //     cout << "here"  << endl;

      i = length;
      while( i-- ) {
      
	//       variables[i]->print(cout);
	//       cout << " " << (variables[i]->isGround()) << " " << (variables[i]->branch) << " " << (variables[i]->getType()) << " =?= " << VariableInt::BOOL << endl;

	if( !variables[i]->isGround() && !variables[i]->branch ) {
	  if( variables[i]->getType() != VariableInt::INTLIST ) {

	    // 	  variables[i]->print( cout );
	    // 	  cout << ": lex val" << endl;

	    variables[i]->branch = new ValSelectorMin( variables[i] );
	  } else {
	    variables[i]->branch = new ValSelectorFirst( variables[i] );
	  }
	}
      }
    } else if(Val == "AntiLex") {
      i = length;
      while( i-- )
	if( !variables[i]->isGround() && !variables[i]->branch ) {
	  if( variables[i]->getType() != VariableInt::INTLIST ) {
	    variables[i]->branch = new ValSelectorMax( variables[i] );
	  } else {
	    variables[i]->branch = new ValSelectorFirst( variables[i] );
	  }
	}
    } else if(Val == "Random") {
      i = length;
      while( i-- )
	if( !variables[i]->isGround() && !variables[i]->branch ) {
	  if( variables[i]->getType() == VariableInt::RANGE ) {
	    variables[i]->branch = new ValSelectorRandMinMax( variables[i] );
	  } else if( variables[i]->getType() != VariableInt::INTLIST ) {

	    // 	  variables[i]->print( cout );
	    // 	  cout << ": rand val" << endl;

	    variables[i]->branch = new ValSelectorRand( variables[i] );
	  } else {
	    variables[i]->branch = new ValSelectorFirst( variables[i] );
	  }
	}
    } else if(Val == "RandomMinMax") {
      i = length;
      while( i-- )
	if( !variables[i]->isGround() && !variables[i]->branch ) {
	  if( variables[i]->getType() != VariableInt::INTLIST ) {
	    variables[i]->branch = new ValSelectorRandMinMax( variables[i] );
	  } else {
	    variables[i]->branch = new ValSelectorFirst( variables[i] );
	  }
	}
    } else if(Val == "DomainSplit") {
      i = length;
      while( i-- )
	if( !variables[i]->isGround() && !variables[i]->branch ) {	
	  if( variables[i]->getType() == VariableInt::INTLIST ) {
	    variables[i]->branch = new ValSelectorFirst( variables[i] );
	  } else if( variables[i]->domsize() > 3 && variables[i]->getType() != VariableInt::LIST )
	    variables[i]->branch = new ValSelectorSplit( variables[i] );
	} else {
	  variables[i]->branch = new ValSelectorMin( variables[i] );
	} 
    }
  }
}



void Solver::setRandomValueOrdering() {
  int i = length;
  while( i-- )
    if( !variables[i]->isGround() && !variables[i]->branch ) {
      if( variables[i]->getType() == VariableInt::RANGE ) {
	variables[i]->branch = new ValSelectorRandMinMax( variables[i] );
      } else if( variables[i]->getType() != VariableInt::INTLIST ) {
	variables[i]->branch = new ValSelectorRand( variables[i] );
      } else {
	variables[i]->branch = new ValSelectorFirst( variables[i] );
      }
    }
}

void Solver::setGuidedOrdering(VarArray& scope, int* ideal) {
  BuildObject *bvar;
  VariableInt *x;
  for(int i=0; i<scope.size(); ++i) {
    bvar = scope[i].var_ptr_;
    if( bvar->isSearchable() && 
	((x = bvar->getVariable()) != NULL) && 
	(x->getType() == VariableInt::BOOL) )
      x->branch = new ValSelectorGuided( x, ideal[i] );
  }
}

void Solver::setRandGuidedOrdering(VarArray& scope, int* ideal, int* proba, int* range) {
  BuildObject *bvar;
  VariableInt *x;
  for(int i=0; i<scope.size(); ++i) {
    bvar = scope[i].var_ptr_;
    if( bvar->isSearchable() && 
	((x = bvar->getVariable()) != NULL) && 
	(x->getType() == VariableInt::BOOL) )
      x->branch = new ValSelectorRandGuided( x, ideal[i], proba[i], range[i] );
  }
}
